import unittest

from sigmaepsilon.deepdict import DeepDict


class TestCompliance(unittest.TestCase):
    def setUp(self):
        self.custom_dict = DeepDict()

    def test_set_and_get_item(self):
        # Test setting and getting items
        self.custom_dict["key1"] = "value1"
        self.assertEqual(self.custom_dict["key1"], "value1")

    def test_len(self):
        # Test the len() function
        self.custom_dict["key1"] = "value1"
        self.custom_dict["key2"] = "value2"
        self.assertEqual(len(self.custom_dict), 2)

    def test_keys(self):
        # Test the keys() method
        self.custom_dict["key1"] = "value1"
        self.custom_dict["key2"] = "value2"
        self.assertEqual(list(self.custom_dict.keys()), ["key1", "key2"])

    def test_values(self):
        # Test the values() method
        self.custom_dict["key1"] = "value1"
        self.custom_dict["key2"] = "value2"
        self.assertEqual(list(self.custom_dict.values()), ["value1", "value2"])

    def test_items(self):
        # Test the items() method
        self.custom_dict["key1"] = "value1"
        self.custom_dict["key2"] = "value2"
        self.assertEqual(
            list(self.custom_dict.items()), [("key1", "value1"), ("key2", "value2")]
        )

    def test_contains(self):
        # Test the 'in' operator
        self.custom_dict["key1"] = "value1"
        self.assertTrue("key1" in self.custom_dict)
        self.assertFalse("key2" in self.custom_dict)

    def test_delete_item(self):
        # Test deleting items
        self.custom_dict["key1"] = "value1"
        del self.custom_dict["key1"]
        self.assertNotIn("key1", self.custom_dict)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import unittest

from sigmaepsilon.core.testing import SigmaEpsilonTestCase
from sigmaepsilon.core.decorate import suppress
from sigmaepsilon.deepdict import DeepDict, parseaddress, asciiprint


class TestPickle(SigmaEpsilonTestCase):
    def test_utils(self):
        data = DeepDict()
        self.assertFailsProperly(ValueError, parseaddress, [], "a")
        self.assertFailsProperly(KeyError, parseaddress, data, "a")

    def test_ascii(self):
        d = {
            "a": {"aa": 1},
            "b": 2,
            "c": {"cc": {"ccc": 3}},
        }
        data = DeepDict.wrap(d)
        suppress(asciiprint)(data)

        self.assertFailsProperly(TypeError, asciiprint, data, dtype=float)
        self.assertFailsProperly(TypeError, asciiprint, 10)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import unittest

from sigmaepsilon.core.testing import SigmaEpsilonTestCase
from sigmaepsilon.deepdict import DeepDict, Key, Value


class TestBehaviour(SigmaEpsilonTestCase):
    def test_behaviour_1(self):
        d = DeepDict()
        d[Key((1, 2))] = "A"
        self.assertEqual(d[Key((1, 2))], "A")
        self.assertFalse((1, 2) in d)
        self.assertTrue(Key((1, 2)) in d)
        del d[Key((1, 2))]
        self.assertFalse(Key((1, 2)) in d)
        d[Key((1, 2))] = "A"
        d[Key((1, 2))] = None

    def test_behaviour_2(self):
        d = DeepDict()
        d[1, 2] = "A"
        self.assertTrue((1, 2) in d)

    def test_behaviour_3(self):
        data = DeepDict()
        data["a", "b", "c", "e"] = 1
        self.assertEqual(data["a", "b", "c"].parent.key, "b")
        self.assertFalse(data["a", "b", "c"].is_root())
        self.assertTrue(data["a", "b", "c"].is_leaf())

    def test_behaviour_4(self):
        data = DeepDict()
        data["a", "b", "c", "e"] = 1
        data.lock()
        self.assertTrue(data.locked)

        def foo():
            data["b"] = 1

        self.assertFailsProperly(KeyError, foo)

    def test_behaviour_5(self):
        d = {
            "a": {"aa": 1},
            "b": 2,
            "c": {"cc": {"ccc": 3}},
        }

        def foo():
            DeepDict(d)["c", "cc", "ccc"]

        self.assertFailsProperly(AttributeError, foo)

    def test_behaviour_6(self):
        data = DeepDict()
        data["a", "b", "c", "e"] = Value(1)
        self.assertEqual(data["a", "b", "c"].parent.key, "b")
        self.assertFalse(data["a", "b", "c"].is_root())
        self.assertTrue(data["a", "b", "c"].is_leaf())

    def test_behaviour_7(self):
        data = DeepDict(a=1, b=DeepDict(c=2))
        data["d"] = 2
        self.assertEqual(data["d"], 2)
        self.assertEqual(data["b", "c"], 2)
        self.assertEqual(data["b"].address[0], "b")
        self.assertEqual(len(data.address), 0)

    def test_behaviour_8(self):
        data = DeepDict(a=1)

        def foo():
            [] in data

        self.assertFailsProperly(ValueError, foo)

    def test_behaviour_9(self):
        data = DeepDict()
        data["a", "b", "c", "e"] = 1
        data["a"]["b"]["d"] = 2
        b = data["a", "b"]
        b["e"] = 3
        b["f"] = 1, 2, 3

        for _ in data.keys(deep=True, return_address=True):
            pass

        for _ in data.values(deep=True, return_address=True):
            pass

        for _ in data.items(deep=True, return_address=True):
            pass

        for _ in data.keys(deep=True, return_address=False):
            pass

        for _ in data.values(deep=True, return_address=False):
            pass

        for _ in data.items(deep=True, return_address=False):
            pass

    def test_behaviour_10(self):
        data = DeepDict()
        data["a", "b", "c"] = 1
        self.assertTrue(data.name is None)
        self.assertEqual(data["a"].name, "a")

        data.name = "data"
        self.assertEqual(data.name, "data")
        self.assertFailsProperly(TypeError, lambda v: setattr(data, "name", v), 10)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import unittest

from sigmaepsilon.deepdict import DeepDict


class TestDeepDict(unittest.TestCase):
    def test_deepdict(self):
        """
        Qualitative test on basic usage of a DeepDict.
        """
        data = DeepDict()
        data["a", "b", "c"] = 1
        f = data.containers
        assert [c.key for c in f(inclusive=True, deep=True)] == [None, "a", "b"]
        assert [c.key for c in f(inclusive=True, deep=False)] == [None, "a"]
        assert [c.key for c in f(inclusive=False, deep=True)] == ["a", "b"]
        assert [c.key for c in f(inclusive=False, deep=False)] == ["a"]
        assert data["a", "b"].depth == 2
        assert data["a", "b"].key == "b"
        assert data.is_root() is True
        assert data["a", "b"].is_root() is False
        assert data["a", "b"].parent == data["a"]
        assert data["a", "b"].root == data

        # lock test
        has_error = False
        data.lock()
        assert data.locked
        try:
            data["a", "b", "d"] = 2
        except:
            has_error = True
        finally:
            assert has_error
        data.unlock()
        assert not data.locked
        has_error = False
        try:
            data["a", "b", "d"] = 2
        except:
            has_error = True
        finally:
            assert not has_error

        # try indexing
        data["a"]["b"]["e"] = 3
        assert data["a", "b", "e"] == 3

        # other stuff
        data["a", "b"].root
        data.__repr__()

    def test_contains(self):
        data = DeepDict()
        data["a", "b", "c"] = 1
        self.assertTrue(["a", "b", "c"] in data)
        self.assertTrue("a" in data)

    def test_lib_compliance(self):
        """
        Tests to assure that a `DeepDict` works the same way as a `dict`.
        """
        data1 = DeepDict(a=1, b=dict(c=2, d=3))
        data2 = dict(a=1, b=dict(c=2, d=3))
        assert data1 == data2
        assert list(data1.keys()) == list(data1.keys())
        assert list(data1.values()) == list(data1.values())


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from typing import Iterable, Tuple, Any, List, Hashable, Optional, Union, TypeVar, Callable
from copy import copy

try:
    import asciitree
except ImportError:
    asciitree = None

__all__ = ["dictparser", "parseaddress", "parseitems", "parsedicts", "parsedicts_addr"]


DictLike = TypeVar("DictLike")


def dictparser(d: dict, *, dtype=dict, **_kw) -> Iterable[Tuple[List[Hashable], Any]]:
    """
    Iterates through all the values of a nested dictionary.

    Notes
    -----
    Returns all kinds of items, even nested discionaries themselves,
    along with their content.

    Example
    -------
    >>> from sigmaepsilon.deepdict import dictparser
    >>> d = {
    ...     "a" : {"aa" : 1},
    ...     "b" : 2,
    ...     "c" : {"cc" : {"ccc" : 3}},
    ... }
    >>> for address, value in dictparser(d):
    ...     print(f"Address: {address}, Value: {value}")
    Address: ['a', 'aa'], Value: 1
    Address: ['b'], Value: 2
    Address: ['c', 'cc', 'ccc'], Value: 3
    """
    address = _kw.get("_addr", [])
    for key, value in d.items():
        subaddress = copy(address)
        subaddress.append(key)
        if isinstance(value, dtype):
            for data in dictparser(value, dtype=dtype, _addr=subaddress):
                yield data
        else:
            yield subaddress, value


def parseaddress(d: dict, address: List[Hashable]) -> Any:
    """
    Returns a value specified with an address.

    Example
    -------
    >>> from sigmaepsilon.deepdict import parseaddress
    >>> d = {
    ...     "a" : {"aa" : 1},
    ...     "b" : 2,
    ...     "c" : {"cc" : {"ccc" : 3}},
    ... }
    >>> parseaddress(d, ['c', 'cc', 'ccc'])
    3
    """
    if not isinstance(d, dict):
        raise ValueError

    if not address[0] in d:
        raise KeyError(address[0])

    if len(address) > 1:
        return parseaddress(d[address[0]], address[1:])
    else:
        return d[address[0]]


def parseitems(d: dict, *, dtype: Any = dict) -> Iterable[Tuple[Hashable, Any]]:
    """
    A generator function that yields all the items of a nested dictionary as
    (key, value) pairs.

    Notes
    -----
    Does not return nested dictionaries themselves, only their content.

    Example
    -------
    >>> from sigmaepsilon.deepdict import parseitems
    >>> d = {
    ...     "a" : {"aa" : 1},
    ...     "b" : 2,
    ...     "c" : {"cc" : {"ccc" : 3}},
    ... }
    >>> for key, value in parseitems(d):
    ...     print(f"Key: {key}, Value: {value}")
    Key: aa, Value: 1
    Key: b, Value: 2
    Key: ccc, Value: 3
    """
    for key, value in d.items():
        if isinstance(value, dtype):
            for data in parseitems(value, dtype=dtype):
                yield data
        else:
            yield key, value


def parsedicts(
    d: dict, *, inclusive: bool = True, dtype: Any = dict, deep: bool = True
) -> Iterable[dict]:
    """
    Returns all subdirectories of a dictionary.

    Example
    -------
    >>> from sigmaepsilon.deepdict import parsedicts
    >>> d = {
    ...     "a" : {"aa" : 1},
    ...     "b" : 2,
    ...     "c" : {"cc" : {"ccc" : 3}},
    ... }

    >>> for subd in parsedicts(d):
    ...     print(subd)
    {'a': {'aa': 1}, 'b': 2, 'c': {'cc': {'ccc': 3}}}
    {'aa': 1}
    {'cc': {'ccc': 3}}
    {'ccc': 3}

    >>> for subd in parsedicts(d, inclusive=False):
    ...     print(subd)
    {'aa': 1}
    {'cc': {'ccc': 3}}
    {'ccc': 3}
    """
    if inclusive:
        if isinstance(d, dtype):
            yield d

    for value in d.values():
        if isinstance(value, dtype):
            yield value
            if deep:
                for subvalue in parsedicts(value, inclusive=False, dtype=dtype):
                    yield subvalue


def parsedicts_addr(
    d: dict, *, inclusive: bool = True, dtype: Any = dict, deep: bool = True, **_kw
) -> Tuple[List[Hashable], Iterable[dict]]:
    """
    Returns all subdirectories of a dictionary and their addresses.

    Example
    -------
    >>> from sigmaepsilon.deepdict import parsedicts_addr
    >>> d = {
    ...     "a" : {"aa" : 1},
    ...     "b" : 2,
    ...     "c" : {"cc" : {"ccc" : 3}},
    ... }

    >>> for addr, subd in parsedicts_addr(d):
    ...     print(f"{subd} @ {addr}")
    {'a': {'aa': 1}, 'b': 2, 'c': {'cc': {'ccc': 3}}} @ []
    {'aa': 1} @ ['a']
    {'cc': {'ccc': 3}} @ ['c']
    {'ccc': 3} @ ['c', 'cc']

    >>> for addr, subd in parsedicts_addr(d, inclusive=False):
    ...     print(f"{subd} @ {addr}")
    {'aa': 1} @ ['a']
    {'cc': {'ccc': 3}} @ ['c']
    {'ccc': 3} @ ['c', 'cc']
    """
    address = _kw.get("_addr", [])

    if inclusive:
        if isinstance(d, dtype):
            yield address, d

    for key, value in d.items():
        if isinstance(value, dtype):
            addr = copy(address)
            addr.append(key)
            yield addr, value
            if deep:
                for subaddr, subval in parsedicts_addr(
                    value, inclusive=False, dtype=dtype, _addr=addr
                ):
                    yield subaddr, subval


def _asciitree(data: dict, dtype: type = dict, **_kw) -> dict:
    tree = _kw.get("_tree", {})
    name = getattr(data, "name", data.__class__.__name__)
    name = data.__class__.__name__ if name is None else name
    tree[name] = {}
    for value in data.values():
        if isinstance(value, dtype):
            _asciitree(value, dtype=dtype, _tree=tree[name])
    return tree


def _wrap(data: dict, wrapper: DictLike, tr:Optional[Union[Callable, None]]=None, **_kw) -> DictLike:
    result = _kw.get("_result", None)
    if result is None:
        result = wrapper()
        
    if tr is None:
        tr = lambda x: x
        
    for key, value in data.items():
        if isinstance(value, dict):
            result[key] = wrapper()
            _wrap(value, wrapper, tr=tr, _result=result[key])
        else:
            result[key] = tr(value)
            
    return result


if asciitree is None:  # pragma: no cover

    def asciiprint(*_, **__) -> str:
        raise ImportError("This requires the 'asciitree' package.")


else:

    def asciiprint(
        data: dict,
        *,
        dtype: Optional[Union[type, None]] = None,
        tr: Optional[Union[asciitree.KeyArgsConstructor, None]] = None,
    ) -> None:
        """
        Prints a dictionary as a tree using the ASCII character set.

        Parameters
        ----------
        data: dict, Optional
            A dictionary.
        dtype: type, Optional
            If a valid type is provided (a subclass of `dict`), then the tree
            will only include containers of that class. Default is None, in which case
            only containers with the same type as 'data' are included in the output.
        tr: asciitree.KeyArgsConstructor, Optional
            A formatter to use. Default is None, in which case a `asciitree.LeftAligned`
            instance is created at runtime.

        Notes
        -----
        This requires the `asciitree` package to be installed.

        Example
        -------
        >>> from sigmaepsilon.deepdict import DeepDict, asciiprint
        >>> d = {
        ...     "a" : {"aa" : 1},
        ...     "b" : 2,
        ...     "c" : {"cc" : {"ccc" : 3}},
        ... }
        >>> data = DeepDict.wrap(d)
        >>> asciiprint(data)
        DeepDict
         +-- a
         +-- c
             +-- cc
        """
        dtype = data.__class__ if dtype is None else dtype
        
        if not isinstance(data, dict):
            raise TypeError("Type of 'data' must be a subclass of 'dict'")
        
        if not issubclass(dtype, dict):
            raise TypeError("'dtype' must be a subclass of 'dict'")

        tr = asciitree.LeftAligned() if tr is None else tr

        print(tr(_asciitree(data)))



# -*- coding: utf-8 -*-
import unittest

from sigmaepsilon.deepdict import DeepDict


class TestDelitem(unittest.TestCase):
    def test_delitem_1(self):
        dd = DeepDict()
        dd["A", "B"] = DeepDict()
        del dd["A", "B"]
        keys = list(dd.keys())
        self.assertTrue(len(keys) == 1)
        self.assertEqual(keys[0], "A")

    def test_delitem_2(self):
        from sigmaepsilon.deepdict import DeepDict

        dd = DeepDict()
        dd["A", "B"] = DeepDict()
        del dd["A"]
        keys = list(dd.keys())
        self.assertTrue(len(keys) == 0)


if __name__ == "__main__":
    unittest.main()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# For ideas:

# https://github.com/pradyunsg/furo/blob/main/docs/conf.py
# https://github.com/sphinx-gallery/sphinx-gallery/blob/master/doc/conf.py

# --------------------------------------------------------------------------

import sys
import os
from datetime import date

import sigmaepsilon.deepdict as library

from sphinx.config import Config

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath("../../src"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = library.__pkg_name__
copyright = "2014-%s, Bence Balogh" % date.today().year
author = "Bence Balogh"


def setup(app: Config):
    app.add_config_value("project_name", project, "html")


# The short X.Y version.
version = library.__version__
# The full version, including alpha/beta/rc tags.
release = "v" + library.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    # allows to work with markdown files
    "myst_parser",  # pip install myst-parser for this
    # to plot summary about durations of file generations
    "sphinx.ext.duration",
    # to test code snippets in docstrings
    "sphinx.ext.doctest",
    # for automatic exploration of the source files
    "sphinx.ext.autodoc",
    # to enable cross referencing other documents on the internet
    "sphinx.ext.intersphinx",
    # Napoleon is a extension that enables Sphinx to parse both NumPy and Google style docstrings
    "sphinx.ext.napoleon",
    #'sphinx_gallery.gen_gallery',
    #'sphinx_gallery.load_style',  # load CSS for gallery (needs SG >= 0.6)
    "nbsphinx",  # to handle jupyter notebooks
    # "nbsphinx_link",  # for including notebook files from outside the sphinx source root
    "sphinx_copybutton",  # for "copy to clipboard" buttons
    #"sphinx.ext.mathjax",  # for math equations
    #"sphinxcontrib.bibtex",  # for bibliographic references
    #"sphinxcontrib.rsvgconverter",  # for SVG->PDF conversion in LaTeX output
    #"sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.extlinks",
    "sphinx_design",
    "sphinx_inline_tabs",
]

autosummary_generate = True

templates_path = ["_templates"]

exclude_patterns = ["_build"]

source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}

# The master toctree document.
master_doc = "index"

language = "EN"

# See warnings about bad links
nitpicky = True
nitpick_ignore = [
    ("", "Pygments lexer name 'ipython' is not known"),
    ("", "Pygments lexer name 'ipython3' is not known"),
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"
pygments_dark_style = "github-dark"
highlight_language = "python3"

intersphinx_mapping = {
    "python": (r"https://docs.python.org/{.major}".format(sys.version_info), None),
    "numpy": (r"https://numpy.org/doc/stable/", None),
    "scipy": (r"http://docs.scipy.org/doc/scipy/reference", None),
    "matplotlib": (r"https://matplotlib.org/stable", None),
    "sphinx": (r"https://www.sphinx-doc.org/en/master", None),
    "pandas": (r"https://pandas.pydata.org/pandas-docs/stable/", None),
    "sigmaepsilon.core": (r"https://sigmaepsiloncore.readthedocs.io/en/latest/", None),
    "sigmaepsilon.math": (r"https://sigmaepsilonmath.readthedocs.io/en/latest/", None),
    "linkeddeepdict": (r"https://linkeddeepdict.readthedocs.io/en/latest/", None),
}

# -- bibtex configuration -------------------------------------------------
# https://sphinxcontrib-bibtex.readthedocs.io/en/latest/usage.html

bibtex_bibfiles = ["references.bib"]
bibtex_default_style = "unsrt"

# If no encoding is specified, utf-8-sig is assumed.
# bibtex_encoding = 'latin'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"

html_theme_options = {
    "show_prev_next": True,
    "icon_links": [
        {
            "name": "GitHub",
            "url": f"https://github.com/sigma-epsilon/{project}",
            "icon": "fab fa-github",
            "type": "fontawesome",
        },
        {
            "name": "PyPi",
            "url": f"https://pypi.org/project/{project}/",
            "icon": "fas fa-box-open",
            "type": "fontawesome",
        },
    ],
    "logo": {
        # Because the logo is also a homepage link, including "home" in the alt text is good practice
        "text": f"SigmaEpsilon.DeepDict {version}",
    }
}

html_context = {
   # ...
   "default_mode": "dark"
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- nbsphinx configuration -------------------------------------------------

# This is processed by Jinja2 and inserted before each notebook
nbsphinx_prolog = r"""
{% set docname = "docs\\source\\" + env.doc2path(env.docname, base=None) %}

.. raw:: html

    <div class="admonition note">
      This page was generated from
      <a class="reference external" href="https://github.com/sigma-epsilon/{{ env.config.project_name }}/blob/{{ env.config.release|e }}/{{ docname|e }}">{{ docname|e }}</a>.
    </div>

.. raw:: latex

    \nbsphinxstartnotebook{\scriptsize\noindent\strut
    \textcolor{gray}{The following section was generated from
    \sphinxcode{\sphinxupquote{\strut {{ docname | escape_latex }}}} \dotfill}}
"""

# This is processed by Jinja2 and inserted after each notebook
nbsphinx_epilog = r"""
{% set docname = 'docs/source/' + env.doc2path(env.docname, base=None) %}
.. raw:: latex

    \nbsphinxstopnotebook{\scriptsize\noindent\strut
    \textcolor{gray}{\dotfill\ \sphinxcode{\sphinxupquote{\strut
    {{ docname | escape_latex }}}} ends here.}}
"""



# http://stackoverflow.com/a/6190500/562769
from typing import Hashable, Union, Tuple, Any, TypeVar, Iterable, Optional, Generic
from copy import copy as shallow_copy, deepcopy as deep_copy

from sigmaepsilon.core.typing import issequence
from sigmaepsilon.core import Wrapper

from .utils import dictparser, parseitems, parsedicts, _wrap


__all__ = ["DeepDict", "Key", "Value"]


T = TypeVar("T", bound="DeepDict")


class Key(Wrapper):
    """
    Helper class for keys.
    """

    def __init__(self, arg: Any):
        super().__init__(wrap=arg)


class Value(Wrapper):
    """
    Helper class for values.
    """

    def __init__(self, arg: Any):
        super().__init__(wrap=arg)


class DeepDict(dict, Generic[T]):
    """
    An nested dictionary class with a self-replicating default factory.
    It can be a drop-in replacement for the bulit-in dictionary type,
    but it's more capable as it handles nested layouts.

    Parameters
    ----------
    *args*: tuple, Optional
        Extra positional arguments are forwarded to the `dict` class.
    **kwargs**: dictionary, Optional
        Extra keyword arguments are forwarded to the `dict` class.

    Examples
    --------
    Basic usage:

    >>> from sigmaepsilon.deepdict import DeepDict
    >>> d = {'a' : {'aa' : {'aaa' : 0}}, 'b' : 1, 'c' : {'cc' : 2}}
    >>> dd = DeepDict(d)
    >>> list(dd.values(deep=True))
    [0, 1, 2]
    """

    def __init__(self, *args, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, DeepDict):
                v._parent = self
                v._key = k
        super().__init__(*args, **kwargs)
        self._parent = None
        self._root = None
        self._locked = None
        self._key = None
        self._name = None

    @property
    def parent(self: T) -> Union[T, None, "DeepDict"]:
        """
        Returns the parent of the instance, or None if it has no parent.
        """
        return self._parent

    @property
    def root(self: T) -> Union[T, None, "DeepDict"]:
        """
        Returns the top-level object in a nested layout.
        """
        if self.parent is None:
            return self
        else:
            if self._root is not None:
                return self._root
            else:
                return self.parent.root

    @property
    def name(self) -> Union[str, None]:
        """
        The name of the dictionary. It is used for representation purposes.
        If a name is not specified, the `key` is returned here. Setting a name
        may be important for the top level entity, since it has no key.
        """
        if self._name is None:
            return self._key
        else:
            return self._name

    @name.setter
    def name(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError(f"Name must be a string, it is {type(value)}")
        self._name = value

    @property
    def key(self) -> Union[Hashable, None]:
        """
        Returns the key of the instance, or `None` if it has no parent.
        """
        return self._key

    @property
    def locked(self) -> bool:
        """
        Returns `True` if the object is locked. The property is equpped with a setter.
        """
        if self.parent is None:
            return self._locked if isinstance(self._locked, bool) else False
        else:
            return (
                self._locked if isinstance(self._locked, bool) else self.parent.locked
            )

    @property
    def depth(self) -> int:
        """
        Retuns the depth of the actual instance in a layout, starting from 0..
        """
        if self.parent is None:
            return 0
        else:
            return self.parent.depth + 1

    @property
    def address(self) -> Union[Tuple, None]:
        """Returns the address of an item or `None` it has no parent."""
        if self.is_root():
            return []
        else:
            r = self.parent.address
            r.append(self.key)
            return r

    @classmethod
    def wrap(
        cls: T, d: dict, copy: bool = False, deepcopy: bool = False
    ) -> Union[T, "DeepDict"]:
        """
        Wraps a dictionary with all nested dictionaries and content.

        Parameters
        ----------
        d: dict
            The dictionary to wrap.
        copy: bool, Optional
            If `True`, shallow copies of the values are stored. Default is False.
        deepcopy: bool, Optional
            If `True`, deep copies of the values are stored. Default is False.

        Example
        -------
        >>> from sigmaepsilon.deepdict import DeepDict
        >>> d = {
        ...     "a" : {"aa" : 1},
        ...     "b" : 2,
        ...     "c" : {"cc" : {"ccc" : 3}},
        ... }
        >>> DeepDict.wrap(d)["c", "cc", "ccc"]
        3

        >>> list(DeepDict.wrap(d).items(deep=True))
        [('aa', 1), ('b', 2), ('ccc', 3)]
        """
        if copy and deepcopy:
            raise ValueError("Only one of 'copy' and 'deepcopy' can be True.")

        tr = None

        if copy:
            tr = shallow_copy

        if deepcopy:
            tr = deep_copy

        return _wrap(d, cls, tr=tr)

    def lock(self) -> None:
        """
        Locks the layout of the dictionary. If a `DeepDict` is locked,
        missing keys are handled the same way as they would've been handled
        if it was a ´dict´. Also, setting or deleting items in a locked
        dictionary and not possible and you will experience an error upon trying.
        """
        self._locked = True

    def unlock(self) -> None:
        """
        Releases the layout of the dictionary. If a `DeepDict` is not locked,
        a missing key creates a new level in the layout, also setting and deleting
        items becomes an option.
        """
        self._locked = False

    def is_root(self) -> bool:
        """
        Returns `True`, if the instance is the root.
        """
        return self.parent is None

    def is_leaf(self: T, dtype: Optional[Any] = None) -> bool:
        """
        Returns `True`, if the instance has no children.

        Parameters
        ----------
        dtype: Any, Optional
            It can be a type that controls what is considered as a children.
            It is None by default, which means that only instances of the same
            class are considered children. In this case, a simple `dict` wouldn't
            make it.

        Example
        -------
        >>> from sigmaepsilon.deepdict import DeepDict
        >>> d = {
        ...     "a" : {"aa" : 1},
        ...     "b" : 2,
        ...     "c" : {"cc" : {"ccc" : 3}},
        ... }
        >>> dd = DeepDict.wrap(d)
        >>> dd.is_leaf(), dd["a"].is_leaf()
        (False, True)
        """
        dtype = self.__class__ if dtype is None else dtype
        return not any(list([isinstance(v, dtype) for v in self.values()]))

    def containers(
        self: T,
        *,
        inclusive: Optional[bool] = False,
        deep: Optional[bool] = True,
        dtype: Optional[Any] = None,
    ) -> Iterable[T]:
        """
        Returns all the containers in a nested layout. A dictionary in a nested layout
        is called a container, only if it contains other containers (it is a parent).

        Parameters
        ----------
        inclusive: bool, Optional
            If `True`, the object the call is made upon also gets returned.
            This can be important if you make the call on the root object, which most
            of the time does not hold onto relevant data directly.
            Default is `False`.
        deep: bool, Optional
            If `True` the parser goes into nested dictionaries.
            Default is `True`
        dtype: Any, Optional
            Constrains the type of the returned objects.
            Default is `None`, which means no restriction.

        Returns
        -------
        generator
            Returns a generator object.

        Examples
        --------
        A simple example:

        >>> from sigmaepsilon.deepdict import DeepDict
        >>> data = DeepDict()
        >>> data['a', 'b', 'c'] = 1
        >>> [c.key for c in data.containers()]
        ['a', 'b']

        We can see, that dictionaries 'a' and 'b' are returned as containers, but 'c'
        isn't,  because it is not a parent, there are no deeper levels.

        >>> [c.key for c in data.containers(inclusive=True, deep=True)]
        [None, 'a', 'b']

        >>> [c.key for c in data.containers(inclusive=True, deep=False)]
        [None, 'a']

        >>> [c.key for c in data.containers(inclusive=False, deep=True)]
        ['a', 'b']

        >>> [c.key for c in data.containers(inclusive=False, deep=False)]
        ['a']
        """
        dtype = self.__class__ if dtype is None else dtype
        return parsedicts(self, inclusive=inclusive, dtype=dtype, deep=deep)

    def __getitem__(self: T, key) -> Union[Any, T]:
        try:
            if isinstance(key, Key):
                return super().__getitem__(key.wrapped)
            elif issequence(key):
                item = self.__getitem__(key[0])
                if len(key) > 1:
                    return item.__getitem__(key[1:])
                else:
                    return item
            else:
                return super().__getitem__(key)
        except ValueError:
            return self.__missing__(key)
        except KeyError:
            return self.__missing__(key)

    def __delitem__(self, key) -> None:
        if isinstance(key, Key):
            super().__delitem__(key.wrapped)
        elif issequence(key):
            parent = self.__getitem__(key[:-1])
            parent.__delitem__(key[-1])
        else:
            super().__delitem__(key)

    def __setitem__(self, key, value) -> None:
        try:
            if isinstance(key, Key) or not issequence(key):
                if key in self:
                    d = self[key]
                    if isinstance(d, DeepDict):
                        d.__leave_parent__()
                else:
                    d = self.__missing__(key)

                if value is None:
                    if key in self:
                        del self[key]
                else:
                    if isinstance(value, DeepDict):
                        value.__join_parent__(self, key)

                    if isinstance(key, Key):
                        return super().__setitem__(key.wrapped, value)
                    else:
                        return super().__setitem__(key, value)
            else:
                if not key[0] in self:
                    d = self.__missing__(key[0])
                else:
                    d = self[key[0]]

                if len(key) > 1:
                    d.__setitem__(key[1:], value)
                else:
                    d = self[key[0]]
                    if isinstance(d, DeepDict):
                        d.__leave_parent__()

                    if value is None:
                        del self[key[0]]
                    else:
                        self[key[0]] = value

        except KeyError:
            return self.__missing__(key)

    def __missing__(self: T, key) -> T:
        if self.locked:
            raise KeyError(f"Missing key '{key}' and the object is locked!")

        if isinstance(key, Key) or not issequence(key):
            value = self.__class__()
            value.__join_parent__(self, key)
            if isinstance(key, Key):
                super().__setitem__(key.wrapped, value)
            else:
                super().__setitem__(key, value)
            return value
        elif issequence(key):
            if key[0] not in self:
                value = self.__class__()
                value.__join_parent__(self, key)
                super().__setitem__(key[0], value)
            else:
                value = self[key[0]]

            if len(key) > 1:
                return value.__missing__(key[1:])
            else:
                return value

    def __contains__(self, item: Any) -> bool:
        if isinstance(item, Key):
            return super().__contains__(item.wrapped)
        elif issequence(item):
            if len(item) == 0:
                raise ValueError(f"{item} has zero length")
            else:
                obj = self
                for subitem in item:
                    if not isinstance(subitem, Hashable):
                        raise TypeError(f"{subitem} is not hashable")
                    else:
                        if obj.__contains__(subitem):
                            obj = obj.__getitem__(subitem)
                        else:
                            return False
                return True
        elif isinstance(item, Hashable):
            return super().__contains__(item)
        else:
            raise TypeError(f"{item} is not hashable")

    def __reduce__(self) -> Any:
        return self.__class__, tuple(), None, None, self.items()

    def __repr__(self) -> str:
        frmtstr = self.__class__.__name__ + "(%s)"
        return frmtstr % (dict.__repr__(self))

    def __leave_parent__(self) -> None:
        self._parent = None
        self._root = None
        self._key = None

    def __join_parent__(self: T, parent: T, key: Optional[Hashable] = None) -> None:
        self._parent = parent
        self._root = parent.root
        self._key = key

    def items(
        self: T, *, deep: Optional[bool] = False, return_address: Optional[bool] = False
    ) -> Iterable[Tuple]:
        """
        Returns the items. When called without arguments, it works the same as for
        standard dictionaries.

        Parameters
        ----------
        deep: bool, Optional
            If `True` the parser goes into nested dictionaries.
            Default is `True`.
        return_address: bool, Optional
            If `True`, addresses are returned instead of keys. The difference is similar
            than that of absolute and repative paths. In this respect, keys are the relative
            paths (relative to the parent), and addresses are absolute paths (relative to the root).
            Default is False.
        """
        if deep:
            if return_address:
                for addr, v in dictparser(self):
                    yield addr, v
            else:
                for k, v in parseitems(self):
                    yield k, v
        else:
            for k, v in super().items():
                yield k, v

    def values(
        self: T, *, deep: Optional[bool] = False, return_address: Optional[bool] = False
    ) -> Iterable[Union[Any, T]]:
        """
        Returns the values. When called without arguments, it works the same as for
        standard dictionaries.

        Parameters
        ----------
        deep: bool, Optional
            If `True` the parser goes into nested dictionaries.
            Default is `True`.
        return_address: bool, Optional
            If `True`, addresses are returned as well.
            Default is False.
        """
        if deep:
            if return_address:
                for addr, v in dictparser(self):
                    yield addr, v
            else:
                for _, v in parseitems(self):
                    yield v
        else:
            for v in super().values():
                yield v

    def keys(
        self: T, *, deep: Optional[bool] = False, return_address: Optional[bool] = False
    ) -> Iterable[Hashable]:
        """
        Returns the keys. When called without arguments, it works the same as for
        standard dictionaries.

        Parameters
        ----------
        deep: bool, Optional
            If `True` the parser goes into nested dictionaries.
            Default is `True`.
        return_address: bool, Optional
            If `True`, addresses are returned instead of keys. The difference is similar
            than that of absolute and repative paths. In this respect, keys are the relative
            paths (relative to the parent), and addresses are absolute paths (relative to the root).
            Default is False.
        """
        if deep:
            if return_address:
                for addr, _ in dictparser(self):
                    yield addr
            else:
                for k, _ in parseitems(self):
                    yield k
        else:
            for k in super().keys():
                yield k



from os.path import dirname, abspath
from importlib.metadata import metadata

from sigmaepsilon.core.config import namespace_package_name

from .deepdict import DeepDict, Key, Value
from .utils import (
    dictparser,
    parseaddress,
    parsedicts,
    parseitems,
    parsedicts_addr,
    asciiprint,
)

__all__ = [
    "DeepDict",
    "Key",
    "Value",
    "dictparser",
    "parseaddress",
    "parseitems",
    "parsedicts",
    "parsedicts_addr",
    "asciiprint",
]

__pkg_name__ = namespace_package_name(dirname(abspath(__file__)), 10)
__pkg_metadata__ = metadata(__pkg_name__)
__version__ = __pkg_metadata__["version"]
__description__ = __pkg_metadata__["summary"]
del __pkg_metadata__



# -*- coding: utf-8 -*-
import unittest

from sigmaepsilon.core.testing import SigmaEpsilonTestCase
from sigmaepsilon.deepdict import DeepDict


class TestWrap(SigmaEpsilonTestCase):
    def test_wrap(self):
        d = {
            "a": {"aa": 1},
            "b": 2,
            "c": {"cc": {"ccc": 3}},
        }
        self.assertEqual(DeepDict.wrap(d)["a", "aa"], 1)
        self.assertEqual(DeepDict.wrap(d)["b"], 2)
        self.assertEqual(DeepDict.wrap(d)["c", "cc", "ccc"], 3)
        
        DeepDict.wrap(d, copy=True)
        DeepDict.wrap(d, deepcopy=True)
        
        self.assertFailsProperly(ValueError, DeepDict.wrap, d, copy=True, deepcopy=True)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import unittest
import pickle

from sigmaepsilon.core.testing import SigmaEpsilonTestCase
from sigmaepsilon.deepdict import DeepDict


class TestPickle(SigmaEpsilonTestCase):
    def test_pickle(self):
        data = DeepDict()
        data["a", "b", "c", "e"] = 1
        data["a"]["b"]["d"] = 2
        b = data["a", "b"]
        b["e"] = 3
        b["f"] = 1, 2, 3

        serialized_obj = pickle.dumps(data)
        recreated_obj = pickle.loads(serialized_obj)

        self.assertIsInstance(recreated_obj, DeepDict)
        self.assertEqual(recreated_obj["a", "b", "d"], 2)


if __name__ == "__main__":
    unittest.main()
